"""
author: deadc0de6 (https://github.com/deadc0de6)
Copyright (c) 2019, deadc0de6

diverse exceptions
"""


class YamlException(Exception):
    """exception in CfgYaml"""
    pass


class UndefinedException(Exception):
    """exception in templating"""
    pass
